try:
    from urlparse import urljoin
except ImportError:
    # python3 compatibility
    from urllib.parse import urljoin
from zope.dottedname.resolve import resolve


def get_page_url(skin_name, page_mappings, page_id):
    """ Returns the page_url for the given page_id and skin_name """
    fallback = '/'
    if page_id is not None:
        return page_mappings[page_id].get('path', '/')
    return fallback


def get_page_class(skin_name, page_mappings, page_id=None, fallback=None,
                   default_pages=None):
    """ Returns the page class for a given skin name and page mapping.

        First of all, if there is no page id it will return the given fallback
        if defined of the default page for the skin in use.

        If there is a page id, it will return:
        * the match for the given skin if defined
        * a fallback if defined
        * the given fallback if defined or the global default page class
    """
    fallback = fallback and fallback or resolve(default_pages[
        skin_name])
    if not page_id:
        return fallback

    page_class_mapping = page_mappings[page_id].get('page_class', None)
    if page_class_mapping is not None:
        result = page_class_mapping.get(
            skin_name, page_class_mapping.get('fallback', None))
        return result and resolve(result) or fallback

    return fallback


def page_factory(base_url, browser, default_page_class, page_mappings,
                 skin_name, page_id=None, **kwargs):
    url = base_url

    if page_id is None:
        url = base_url
        page_class = default_page_class
    else:
        path = page_mappings[page_id]['path']
        page_class = get_page_class(
            skin_name,
            page_mappings,
            page_id=page_id, fallback=default_page_class)
        url = urljoin(base_url, path)

    page = page_class(browser, base_url=url, **kwargs)

    return page



# -*- coding: utf-8 -*-
"""
The following diagram shows the interactions between the `pytest fixtures`_
created in the ``pypom_navigation`` package:

.. graphviz::

   digraph {
      bdd_vars;
      browser;
      credentials_mapping;
      default_page;
      default_page_class;
      default_pages;
      navigation;
      navigation_class;
      now;
      page_mappings,
      parametrizer;
      parametrizer_class;
      request;
      skin;
      skin_base_url;
      skip_by_skin_names;
      test_run_identifier;
      variables;
      default_timeout -> {variables};
      bdd_vars -> {parametrizer};
      credentials_mapping -> {navigation};
      default_page_class -> {navigation};
      default_pages -> {default_page_class};
      navigation_class -> {navigation};
      now -> {bdd_vars};
      page_mappings -> {default_page_class navigation};
      parametrizer_class -> {parametrizer};
      request -> {skip_by_skin_names};
      skin -> {skin_base_url credentials_mapping default_page_class
               navigation skip_by_skin_names
               test_run_identifier bdd_vars};
      skin_base_url -> {navigation};
      test_run_identifier -> {bdd_vars};
      variables -> {skin_base_url credentials_mapping};
   }


.. _pytest fixtures: http://doc.pytest.org/en/latest/fixture.html
"""

import pytest
import uuid
import datetime

from .util import (
    get_page_class,
)
from .navigation import Navigation
from .parametrizer import Parametrizer


def pytest_configure(config):
    # register an additional marker
    config.addinivalue_line(
        "markers",
        "skip_skins(skins): mark test to be skipped for the given skin ids"
    )


def skip_skins(skins):
    """
    Decorator to mark tests to be skipped for the given skin ids.

    ie. @skip_skins(['skin1', 'skin2'])
    :return pytest.mark:
    """
    return pytest.mark.skip_skins(skins)


@pytest.fixture(scope='session')
def skin():
    """ This fixture provides the skin associated with the application
        on which starts the test session.


        For example:

            @pytest.fixture(scope='session',
                            params=mypackage.DEFAULT_PAGES.keys())
            def skin(request):
                return request.param
    """
    return 'skin1'


@pytest.fixture(scope="session")
def default_pages():
    """ A mapping with the default page object class for each skin

        It's up to you override this fixture with your settings.

        For example::

            DEFAULT_PAGES = {
                'skin1': 'mypackage.pages.BasePage',
            }
    """
    return {'skin1': 'pypom_navigation.pages.BasePage'}


@pytest.fixture(scope="session")
def page_mappings():
    """
        Returns the page mappings that describes for each page id
        info like the page path, the page object class to be used
        or any other information::

            PAGE_MAPPINGS = {
                'HomePage': {'path': '/'},
                'LoginPage': {'path': '/'},
            }

        It's up to you override this fixture with your settings.

        :return: dictionary with all known pages
        :rtype: dict`
    """
    return {}


@pytest.fixture(scope='session')
def skin_base_url(skin, variables):
    """ Returns the skin_base_url associated to the skin.
    """
    return variables['skins'][skin]['base_url']


@pytest.fixture(scope='session')
def credentials_mapping(skin, variables):
    """
        This fixture provides users credentials via a file specified on the
        --variables option. The file format is one supported by
        pytest-variables.

        :return: credentials mapping dictionary with all available credentials
        :rtype: dict
        :raises: KeyError
    """
    return variables['skins'][skin]['credentials']


@pytest.fixture
def default_page_class(skin, page_mappings, default_pages):
    """
        Returns the default page object base class.

        :return: base page object class
        :rtype: :py:class:`tierra_qa.pages.BasePage`
    """
    return get_page_class(
        skin,
        page_mappings,
        default_pages=default_pages,
    )


@pytest.fixture
def default_timeout(variables):
    """ Default page timeout """
    return variables.get('default_timeout', 10)


@pytest.fixture
def navigation(navigation_class,
               default_page_class,
               page_mappings,
               credentials_mapping,
               skin,
               skin_base_url,
               request,
               variables,
               default_timeout):
    """ Wraps a page and a page mappings accessible by
        pages.

        ``navigation.page`` is meant to be mutable since
        through the BDD steps the page instance could
        change.
    """
    nav = navigation_class(
        None,      # backwards compatibility
        default_page_class,
        page_mappings,
        credentials_mapping,
        skin,
        skin_base_url,
        request,
        variables,
        timeout=default_timeout)
    return nav


@pytest.fixture
def navigation_class():
    """ Returns the navigation class used for wrap pages"""

    return Navigation


@pytest.fixture(autouse=True)
def skip_by_skin_names(request, skin):
    """ Skip by skin name.

        We support validation for multi skin applications providing the best
        page object class match.

        We expect many failures we want to avoid because many tests will fail
        because the related page object implementation still not exists.

        If you want you can omit a test execution for a given skin adding a
        a ```@pytest.mark.skip_skins(['skin2'])``` decorator on your tests.

        Tests marked with a skin2 skip will be executed for all skins
        except for skin2.

        See http://bit.ly/2dYnOSv for further info.
    """
    if request.node.get_marker('skip_skins'):
        if skin in request.node.get_marker('skip_skins').args[0]:
            pytest.skip('skipped on this skin: {}'.format(skin))


@pytest.fixture
def test_run_identifier(skin):
    """ Return a session based random prefixed UUID used for
        identifying data created in this test run.
    """
    return "QA-{0}-{1}".format(str(uuid.uuid1()), skin)


@pytest.fixture
def now():
    """ Now fixture, returns current datetime object """
    return datetime.datetime.now()


@pytest.fixture
def bdd_vars(test_run_identifier, skin, now):
    """ BDD step vars for test parametrization for dynamic values
        such as test_run_identifier or datetime
    """
    return {'test_run_identifier': test_run_identifier,
            'skin': skin,
            'datetime': now.isoformat()}


@pytest.fixture
def parametrizer_class():
    """ Provides a parametrizer class used for convert parametrized
        json values to regular python dicts.
    """
    return Parametrizer


@pytest.fixture
def parametrizer(parametrizer_class, bdd_vars):
    """ Parametrizer object """
    return Parametrizer(bdd_vars)



# -*- coding: utf-8 -*-

import json
from string import Template


class Parametrizer(object):
    """ This class let you parametrize your strings
        and convert them to regular Python dictionaries.

        It supports also json.

        Let's try with a **matching name**

        >>> value = '{"baudrate": $baudrate_value}'
        >>> mapping = {"baudrate_value": 250, "name": "a name"}
        >>> parametrizer = Parametrizer(mapping)

        With the ``parametrize`` method you'll get a parametrized
        string:

        >>> parametrizer.parametrize(value)
        '{"baudrate": 250}'


        With the ``json_loads`` method you'll get a parametrized regular
        Python mapping:

        >>> parametrizer.json_loads(value) == {'baudrate': 250}
        True

        And now with **non matching names**

        >>> value = '{"name": "$a_name"}'
        >>> mapping = {"name": "a name"}
        >>> parametrizer = Parametrizer(mapping)

        With the ``parametrize`` method you'll get a parametrized
        string:

        >>> parametrizer.parametrize(value)
        '{"name": "$a_name"}'


        With the ``json_loads`` method you'll get a parametrized regular
        Python mapping:

        >>> parametrizer.json_loads(value) == {'name': '$a_name'}
        True

        And **json not valid**

        >>> value = '{"name": $name}'
        >>> mapping = {"name": "a name"}
        >>> parametrizer = Parametrizer(mapping)

        With the ``parametrize`` method you'll get a parametrized
        string:

        >>> parametrizer.parametrize(value)
        '{"name": a name}'


        Depending on Python version 2 vs 3 you will get a different exception:

        * json.decoder.JSONDecodeError: Expecting value: ...
        * ValueError: No JSON object could be decoded

        >>> import pytest
        >>> with pytest.raises(Exception):
        ...     parametrizer.json_loads(value)
    """

    def __init__(self, mapping):
        self.mapping = mapping

    def parametrize(self, value):
        """ Return the value with template substitution """
        template = Template(value)
        return template.safe_substitute(**self.mapping)

    def json_loads(self, value):
        """ Return the json load of template substitution """
        parametrized = self.parametrize(value)
        return json.loads(parametrized)



try:
    from urlparse import urljoin
except ImportError:
    # python3 compatibility
    from urllib.parse import urljoin

from .util import (
    get_page_class,
    get_page_url,
)


class Navigation(object):
    page_id = None

    def __init__(self,
                 page,
                 default_page_class,
                 page_mappings,
                 credentials_mapping,
                 skin,
                 skin_base_url,
                 request,
                 variables,
                 **kwargs):
        self.setPage(page)
        self.default_page_class = default_page_class
        self.page_mappings = page_mappings
        self.credentials_mapping = credentials_mapping
        self.skin = skin
        self.skin_base_url = skin_base_url
        self.request = request
        self.variables = variables
        self.kwargs = kwargs

    def setPage(self, page, page_id=None):
        """ Set wrapping page and update reference links
            for page and navigation
        """
        self.page = page
        self.page_id = page_id
        if page is not None:
            page.navigation = self

    def merge_kwargs(self, keyword_args):
        """ merge keyword args with default keywordargs """
        kwargs = self.kwargs.copy()
        new_kwargs = keyword_args.copy()
        kwargs.update(new_kwargs)
        return kwargs

    def visit_page(self, page_id, **kwargs):
        """ Visit page id reference in navigation
            class
        """
        kwargs = self.merge_kwargs(kwargs)
        page_url = urljoin(self.skin_base_url, self.get_page_url(page_id))
        page_instance = self.get_page_instance(page_id=page_id, **kwargs)
        page_instance.driver.visit(page_url)
        page_instance.wait_for_page_to_load()
        self.setPage(page_instance, page_id=page_id)
        return page_instance

    def update_page(self, page_id, **kwargs):
        """ Update the wrapped page with the appropriate instance
            mapped to the passed page_id
        """
        kwargs = self.merge_kwargs(kwargs)
        page_instance = self.get_page_instance(page_id=page_id, **kwargs)
        page_instance.wait_for_page_to_load()
        self.setPage(page_instance, page_id=page_id)
        return page_instance

    def action_performed(self, action, fallback=None, **kwargs):
        """ Update the wrapped page with the appropriate instance
            referenced by the given action on the current page
        """
        kwargs = self.merge_kwargs(kwargs)
        page_mapping = self.page_mappings[self.page_id]
        page_id = page_mapping.get('actions', {}).get(action)
        if page_id:
            return self.update_page(page_id, **kwargs)
        else:
            page_instance = self.get_page_instance(fallback=fallback, **kwargs)
            page_instance.wait_for_page_to_load()
            self.setPage(page_instance)
            return page_instance

    def get_page_url(self, page_id):
        """ Return the page url for the current wrapped page """
        return get_page_url(self.skin,
                            self.page_mappings, page_id)

    def get_credentials(self, user_id):
        """ Return a tuple with username and password for the given
            user_id
        """
        user_credentials = self.credentials_mapping[user_id]
        return user_credentials['username'], user_credentials['password']

    def get_page_class(self, page_id=None, fallback=None):
        """ Return the page class """
        fallback = fallback and fallback or self.default_page_class
        return get_page_class(
            self.skin,
            self.page_mappings,
            page_id=page_id,
            fallback=fallback)

    def get_page_instance(self, page_id=None, fallback=None, **kwargs):
        """ Get a fresh page instance """
        page_class = self.get_page_class(page_id=page_id, fallback=fallback)
        return page_class(self.driver, **kwargs)

    @property
    def driver(self):
        """ Return page driver or a fresh browser fixture to be used as driver
            for first initialization
        """
        if self.page is not None:
            value = self.page.driver
        else:
            value = self.request.getfixturevalue('browser')
        return value
